@@ -80,6 +80,7 @@ typedef enum {
     CONNECT_STEP_FIRST,
     CONNECT_STEP_CHECK_RUNNING,
     CONNECT_STEP_STACK_INIT,
+    CONNECT_STEP_CHECK_NSS,
     CONNECT_STEP_IPSEC_START,
     CONNECT_STEP_WAIT_READY,
     CONNECT_STEP_CONFIG_ADD,
@@ -358,7 +359,7 @@ find_helper_libexec (const char *progname, GError **error)
 		PREFIX "/libexec/ipsec/",
 		PREFIX "/lib/ipsec/",
 		"/usr/libexec/ipsec/",
-		"/usr/local/libexec/ipsec/"
+		"/usr/local/libexec/ipsec/",
 		"/usr/lib/ipsec/",
 		"/usr/local/lib/ipsec/",
 		NULL,
@@ -512,6 +513,7 @@ retry_cb (gpointer user_data)
 
 	if (!connect_step (self, &error))
 		connect_failed (self, error, NM_VPN_PLUGIN_FAILURE_CONNECT_FAILED);
+	priv->retries--;
 	g_clear_error (&error);
 
 	return FALSE;
@@ -526,6 +528,9 @@ child_watch_cb (GPid pid, gint status, gpointer user_data)
 	GError *error = NULL;
 	gboolean success;
 
+	DEBUG ("Spawn: child %d exited", pid);
+	unblock_quit (self);
+
 	if (priv->watch_id == 0 || priv->pid != pid) {
 		/* Reap old child */
 		waitpid (pid, NULL, WNOHANG);
@@ -535,9 +540,6 @@ child_watch_cb (GPid pid, gint status, gpointer user_data)
 	priv->watch_id = 0;
 	priv->pid = 0;
 
-	DEBUG ("Spawn: child %d exited", pid);
-	unblock_quit (self);
-
 	if (WIFEXITED (status)) {
 		ret = WEXITSTATUS (status);
 		if (ret)
@@ -553,18 +555,24 @@ child_watch_cb (GPid pid, gint status, gpointer user_data)
 		return;
 	}
 
+	/* Ready step can return a failure even if libreswan is ready,
+	 * but failed to listen to some interfaces due to a bug in older
+	 * libreswan versions. */
 	if (priv->connect_step == CONNECT_STEP_WAIT_READY)
 		success = (ret != 1);
 	else
 		success = (ret == 0);
 
+	/* Ignore failures here, maybe the libreswan daemon is too old. */
+	if (priv->connect_step == CONNECT_STEP_CHECK_NSS)
+		success = TRUE;
+
 	if (success) {
 		/* Success; do the next connect step */
 		priv->connect_step++;
 		priv->retries = 0;
 		success = connect_step (self, &error);
 	} else if (priv->retries) {
-		priv->retries--;
 		g_message ("Spawn: %d more tries...", priv->retries);
 		priv->retry_id = g_timeout_add (100, retry_cb, self);
 		return;
@@ -706,14 +714,95 @@ spawn_pty (NMLibreswanPlugin *self,
            const char *progname,
            ...)
 {
-	int pty_master_fd, md;
+	int pty_master_fd = 0;
 	int stdout_pipe[2], stderr_pipe[2];
 	pid_t child_pid;
 	struct termios termios_flags;
 	va_list ap;
 	GPtrArray *argv;
 	char *cmdline, *arg;
+	int ret;
 
+	/* The pipes */
+	if (pipe (stdout_pipe) == -1) {
+		g_set_error (error, NM_VPN_PLUGIN_ERROR, NM_VPN_PLUGIN_ERROR_LAUNCH_FAILED,
+		             "PTY spawn: failed to create a stdout pipe (%d): %s",
+		             errno, g_strerror (errno));
+		return FALSE;
+	}
+	if (pipe (stderr_pipe) == -1) {
+		g_set_error (error, NM_VPN_PLUGIN_ERROR, NM_VPN_PLUGIN_ERROR_LAUNCH_FAILED,
+		             "PTY spawn: failed to create a stderr pipe (%d): %s",
+		             errno, g_strerror (errno));
+		close (stdout_pipe[0]);
+		close (stdout_pipe[1]);
+		return FALSE;
+	}
+
+	/* Set the parent pipes non-blocking, so we can read big buffers
+	 * in the callback without having to use FIONREAD
+	 * to make sure the callback doesn't block.
+	 */
+	ret = fcntl (stdout_pipe[0], F_GETFL);
+	if (ret == -1) {
+		g_set_error (error, NM_VPN_PLUGIN_ERROR, NM_VPN_PLUGIN_ERROR_LAUNCH_FAILED,
+		             "PTY spawn: F_GETFL on stdout failed (%d): %s",
+		             errno, g_strerror (errno));
+		goto badpipes;
+	}
+	ret = fcntl (stdout_pipe[0], F_SETFL, O_NONBLOCK | ret);
+	if (ret == -1) {
+		g_set_error (error, NM_VPN_PLUGIN_ERROR, NM_VPN_PLUGIN_ERROR_LAUNCH_FAILED,
+		             "PTY spawn: F_SETFL on stdout failed (%d): %s",
+		             errno, g_strerror (errno));
+		goto badpipes;
+	}
+	ret = fcntl (stderr_pipe[0], F_GETFL);
+	if (ret == -1) {
+		g_set_error (error, NM_VPN_PLUGIN_ERROR, NM_VPN_PLUGIN_ERROR_LAUNCH_FAILED,
+		             "PTY spawn: F_GETFL on stderr failed (%d): %s",
+		             errno, g_strerror (errno));
+		goto badpipes;
+	}
+	ret = fcntl (stderr_pipe[0], F_SETFL, O_NONBLOCK | ret);
+	if (ret == -1) {
+		g_set_error (error, NM_VPN_PLUGIN_ERROR, NM_VPN_PLUGIN_ERROR_LAUNCH_FAILED,
+		             "PTY spawn: F_SETFL on stderr failed (%d): %s",
+		             errno, g_strerror (errno));
+		goto badpipes;
+	}
+
+	/* Disable buffering of child writes */
+	ret = fcntl (stdout_pipe[1], F_GETFL);
+	if (ret == -1) {
+		g_set_error (error, NM_VPN_PLUGIN_ERROR, NM_VPN_PLUGIN_ERROR_LAUNCH_FAILED,
+		             "PTY spawn: F_GETFL on child stdout failed (%d): %s",
+		             errno, g_strerror (errno));
+		goto badpipes;
+	}
+	ret = fcntl (stdout_pipe[1], F_SETFL, O_SYNC | ret);
+	if (ret == -1) {
+		g_set_error (error, NM_VPN_PLUGIN_ERROR, NM_VPN_PLUGIN_ERROR_LAUNCH_FAILED,
+		             "PTY spawn: F_SETFL on child stdout failed (%d): %s",
+		             errno, g_strerror (errno));
+		goto badpipes;
+	}
+	ret = fcntl (stderr_pipe[1], F_GETFL);
+	if (ret == -1) {
+		g_set_error (error, NM_VPN_PLUGIN_ERROR, NM_VPN_PLUGIN_ERROR_LAUNCH_FAILED,
+		             "PTY spawn: F_GETFL on child stderr failed (%d): %s",
+		             errno, g_strerror (errno));
+		goto badpipes;
+	}
+	ret = fcntl (stderr_pipe[1], F_SETFL, O_SYNC | ret);
+	if (ret == -1) {
+		g_set_error (error, NM_VPN_PLUGIN_ERROR, NM_VPN_PLUGIN_ERROR_LAUNCH_FAILED,
+		             "PTY spawn: F_SETFL on child stderr failed (%d): %s",
+		             errno, g_strerror (errno));
+		goto badpipes;
+	}
+
+	/* The command line arguments */
 	argv = g_ptr_array_sized_new (10);
 	g_ptr_array_add (argv, (char *) progname);
 
@@ -729,38 +818,39 @@ spawn_pty (NMLibreswanPlugin *self,
 		g_free (cmdline);
 	}
 
-	/* The pipes */
-	pipe (stderr_pipe);
-	pipe (stdout_pipe);
-
 	/* Fork the command */
 	child_pid = forkpty (&pty_master_fd, NULL, NULL, NULL);
 	if (child_pid == 0) {
 		/* in the child */
 
-		close (2);
-		dup (stderr_pipe[1]);
-		close (1);
-		dup (stdout_pipe[1]);
+		if (dup2 (stdout_pipe[1], 1) == -1) {
+			g_error ("PTY spawn: cannot dup stdout (%d): %s.", errno, g_strerror (errno));
+			_exit (-1);
+		}
+		if (dup2 (stderr_pipe[1], 2) == -1) {
+			g_error ("PTY spawn: cannot dup stderr (%d): %s.", errno, g_strerror (errno));
+			_exit (-1);
+		}
 
 		/* Close unnecessary pipes */
 		close (stderr_pipe[0]);
 		close (stdout_pipe[0]);
 
-		if ((md = fcntl (stdout_pipe[1], F_GETFL)) != -1)
-			fcntl (stdout_pipe[1], F_SETFL, O_SYNC | md);
-		if ((md = fcntl (stderr_pipe[1], F_GETFL)) != -1)
-			fcntl (stderr_pipe[1], F_SETFL, O_SYNC | md);
-
 		/* Ensure output is untranslated */
 		setenv ("LC_ALL", "C", 1);
 		setenv ("LANG", "C", 1);
 
 		execv (argv->pdata[0], (char * const*) argv->pdata);
-		g_error ("PTY spawn: cannot exec '%s'", (char *) argv->pdata[0]);
+
+		/* This is probably a rather futile attempt to produce an error message
+		 * as it goes to the piped stderr. */
+		g_error ("PTY spawn: cannot exec '%s' (%d): %s", (char *) argv->pdata[0],
+		         errno, g_strerror (errno));
 		_exit (-1);
 	}
 
+	DEBUG ("PTY spawn: child process %d", child_pid);
+
 	/* Close child side's pipes */
 	close (stderr_pipe[1]);
 	close (stdout_pipe[1]);
@@ -777,17 +867,6 @@ spawn_pty (NMLibreswanPlugin *self,
 	}
 	g_ptr_array_free (argv, TRUE);
 
-	/*  Set pipes non-blocking, so we can read big buffers
-	 *  in the callback without having to use FIONREAD
-	 *  to make sure the callback doesn't block.
-	 */
-	if ((md = fcntl (stdout_pipe[0], F_GETFL)) != -1)
-		fcntl (stdout_pipe[0], F_SETFL, O_NONBLOCK | md);
-	if ((md = fcntl (stderr_pipe[0], F_GETFL)) != -1)
-		fcntl (stderr_pipe[0], F_SETFL, O_NONBLOCK | md);
-	if ((md = fcntl (pty_master_fd, F_GETFL)) != -1)
-		fcntl (pty_master_fd, F_SETFL, O_NONBLOCK | md);
-
 	tcgetattr (pty_master_fd, &termios_flags);
 	cfmakeraw (&termios_flags);
 	cfsetospeed (&termios_flags, __MAX_BAUD);
@@ -804,6 +883,13 @@ spawn_pty (NMLibreswanPlugin *self,
 
 	block_quit (self);
 	return TRUE;
+
+badpipes:
+	close (stderr_pipe[0]);
+	close (stdout_pipe[0]);
+	close (stderr_pipe[1]);
+	close (stdout_pipe[1]);
+	return FALSE;
 }
 
 /****************************************************************/
@@ -1459,6 +1545,18 @@ connect_step (NMLibreswanPlugin *self, GError **error)
 		/* fall through */
 		priv->connect_step++;
 
+	case CONNECT_STEP_CHECK_NSS:
+		/* Start the IPsec service */
+		if (!priv->openswan) {
+			success = do_spawn (self, &priv->pid, NULL, NULL, error,
+			                    priv->ipsec_path, "--checknss", NULL);
+			if (success)
+				priv->watch_id = g_child_watch_add (priv->pid, child_watch_cb, self);
+			return success;
+		}
+		/* fall through */
+		priv->connect_step++;
+
 	case CONNECT_STEP_IPSEC_START:
 		/* Start the IPsec service */
 		if (priv->openswan)
